#!/usr/bin/python2 -s

# Copyright 2010-2012 by Carnegie Mellon University

# @OPENSOURCE_HEADER_START@
# Use of the Prism trend script and related source code is subject
# to the terms of the following licenses:
# 
# GNU Public License (GPL) Rights pursuant to Version 2, June 1991
# Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
# 
# NO WARRANTY
# 
# ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
# PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
# PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
# "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
# KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
# LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
# MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
# OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
# SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
# TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
# WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
# LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
# CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
# CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
# DELIVERABLES UNDER THIS LICENSE.
# 
# Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
# Mellon University, its trustees, officers, employees, and agents from 
# all claims or demands made against them (and any related losses, 
# expenses, or attorney's fees) arising out of, or relating to Licensee's 
# and/or its sub licensees' negligent use or willful misuse of or 
# negligent conduct or willful misconduct regarding the Software, 
# facilities, or other rights or assistance granted by Carnegie Mellon 
# University under this License, including, but not limited to, any 
# claims of product liability, personal injury, death, damage to 
# property, or violation of any laws or regulations.
# 
# Carnegie Mellon University Software Engineering Institute authored 
# documents are sponsored by the U.S. Department of Defense under 
# Contract FA8721-05-C-0003. Carnegie Mellon University retains 
# copyrights in all material produced under this contract. The U.S. 
# Government retains a non-exclusive, royalty-free license to publish or 
# reproduce these documents, or allow others to do so, for U.S. 
# Government purposes only pursuant to the copyright license under the 
# contract clause at 252.227.7013.
# @OPENSOURCE_HEADER_END@

from netsa import script
from netsa.data.times import make_datetime
from netsa.data.format import datetime_silk_hour
from netsa.util.shell import pipeline, command, run_parallel
from netsa.data.format import num_prefix
from netsa.sql import db_connect, db_query

from netsa_trend_util import *

chart_support = False
pdf_support = True

try:
    from rayon.miscutils import InvalidColorSpec, parse_color_string
    import rayon.toolbox
    import rayon.backend_cairo
    chart_support = True
    import cairo
    if cairo.version_info < (1, 3):
        pdf_support = False
except:
    pass

try:
    import sqlite
    sqlite_operationalerror = sqlite.OperationalError
except:
    class sqlite_operationalerror(Exception):
        pass

from datetime import timedelta
import os
import os.path
import shlex
import sys
from time import sleep as time_sleep

# Users don't need to know about deprecations.
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

script.set_title("Flow Trends")
script.set_description("""
    Extracts hourly trend data from a SiLK repository and saves byte,
    packet, or flow data in a delimited output file, or plot it in a
    graphical output file.
""")

script.add_file_param("config-path", "Read trend configuration from this file",
                      required=True)
script.add_flow_params(without_params=['type'])

script.add_text_param("values", """
    Which data to produce on the output: 'bytes', 'packets', or
    'flows'.
""", required=False, default='bytes')

script.add_output_file_param("data-path", """
    Optional destination for delimited text output of data.
""", required=False)

if chart_support:
    script.add_text_param("chart-title", """
        Optional chart title (overrides values in the config file)
    """, required=False)
    script.add_output_file_param("png-path", """
        Optional destination for PNG graphical output file.
    """, required=False, mime_type="image/png")
    script.add_output_file_param("svg-path", """
        Optional destination for SVG graphical output file.
    """, required=False, mime_type="image/svg+xml")
    if pdf_support:
        script.add_output_file_param("pdf-path", """
            Optional destination for PDF graphical output file.
        """, required=False, mime_type="application/pdf")
    script.add_output_file_param("ps-path", """
        Optional destination for PostScript graphical output file.
    """, required=False, mime_type="application/postscript")
    script.add_output_dir_param("html-dir-path", """
        Optional directory destination for HTML output file with
        PNG images.
    """, required=False)

script.add_text_param("flow-bin", """
    Optional bin name for which to retrieve flows.  Forces SiLK data
    to be processed even when information is available in the database
    cache.
""", required=False)

script.add_output_file_param("in-flow-path", """
    Destination for incoming flows from 'flow-bin'.
""", required=False)
script.add_output_file_param("out-flow-path", """
    Destination for outgoing flows from 'flow-bin'.
""", required=False)

script.add_text_param("in-types", "SiLK types for incoming data",
                      default_help="in,inweb", expert=True)
script.add_text_param("out-types", "SiLK types for outgoing data",
                      default_help="out,outweb", expert=True)

script.add_flag_param("no-silk-query",
                      "Never pull data from SiLK, even when information "
                      "is missing from the database cache.")
script.add_flag_param("no-database-query",
                      "Do not query data from the database cache, even when "
                      "it is available.",
                      expert=True)

script.add_flag_param("auto-load-database", 
                      "Load data into database cache (missing data only)",
                      expert=True)
script.add_flag_param("load-database", "Load data into database cache",
                      expert=True)
script.add_text_param("database-user", "Username for database access",
                      expert=True)
script.add_text_param("database-password", "Password for database access",
                      expert=True)

def print_warning(out, msg):
    script_name = os.path.basename(sys.argv[0])
    out.write("%s: %s\n" % (script_name, msg))

########################################################################

def run_pipeline(scheme, in_rwfilter, out_rwfilter, pull_params, data):
    # bin name, fifo, file, key
    in_bin_keys = scheme.get_in_bin_keys()
    out_bin_keys = scheme.get_out_bin_keys()
    bin_info = [[bin, in_bin_keys[i], out_bin_keys[i],
                 script.get_temp_dir_file_name(),
                 script.get_temp_dir_file_name(),
                 script.get_temp_dir_file_name(),
                 script.get_temp_dir_file_name()]
                for (i, bin) in enumerate(scheme.get_bins())]
    # Create the FIFOs.
    run_parallel(*[command("mkfifo", f)
                   for (_, _, _, in_fifo, out_fifo, _, _) in bin_info
                   for f in (in_fifo, out_fifo)])
    in_rwfilter_extra = []
    out_pull_params = pull_params
    if not pull_params.get_start_date():
        fifo_in_to_out = script.get_temp_dir_file_name()
        run_parallel([command("mkfifo", fifo_in_to_out)])
        in_rwfilter_extra.extend(["--fail-destination", fifo_in_to_out])
        out_pull_params = script.Flow_params(filenames=[fifo_in_to_out])
    # Set up the pipelines for run_parallel
    # Pipeline that does all of the rwfilters in the "in" cascade
    trend_in_rwfilter_pipeline = pipeline(
        [command(in_rwfilter + in_rwfilter_extra +
                 pull_params.get_argument_list(),
                 ignore_exit_statuses=[])] +
        [bin.get_in_command(in_fifo)
         for (bin, _, _, in_fifo, _, _, _) in bin_info],
        stdin=sys.stdin, stdout='/dev/null')
    # Pipeline that does all of the rwfilters in the "out" cascade
    trend_out_rwfilter_pipeline = pipeline(
        [command(out_rwfilter +
                 out_pull_params.get_argument_list(),
                 ignore_exit_statuses=[])] +
        [bin.get_out_command(out_fifo)
         for (bin, _, _, _, out_fifo, _, _) in bin_info],
        stdout='/dev/null')
    flow_bin = script.get_param("flow-bin")
    flow_bin_in_path = script.get_param("in-flow-path")
    flow_bin_out_path = script.get_param("out-flow-path")
    # One pipeline for each bin: the rwuniq that collects results.
    trend_in_rwcount_pipelines = []
    for (bin, _, _, in_fifo, _, in_file, _) in bin_info:
        if flow_bin_in_path == None or bin.get_name() != flow_bin:
            trend_in_rwcount_pipelines.append(
                pipeline(command("rwuniq", in_fifo,
                                 "--fields=sensor,sTime",
                                 "--bin-time=3600",
                                 "--bytes", "--packets", "--flows",
                                 "--no-titles",
                                 "--delimited=|",
                                 "--output-path=" + in_file)))
        else:
            trend_in_rwcount_pipelines.append(
                pipeline(command("rwuniq", in_fifo,
                                 "--fields=sensor,sTime",
                                 "--bin-time=3600",
                                 "--bytes", "--packets", "--flows",
                                 "--no-titles",
                                 "--delimited=|",
                                 "--copy-input=" + flow_bin_in_path,
                                 "--output-path=" + in_file)))
    trend_out_rwcount_pipelines = []
    for (bin, _, _, _, out_fifo, _, out_file) in bin_info:
        if flow_bin_out_path == None or bin.get_name() != flow_bin:
            trend_out_rwcount_pipelines.append(
                pipeline(command("rwuniq", out_fifo,
                                 "--fields=sensor,sTime",
                                 "--bin-time=3600",
                                 "--bytes", "--packets", "--flows",
                                 "--no-titles",
                                 "--delimited=|",
                                 "--output-path=" + out_file)))
        else:
            trend_out_rwcount_pipelines.append(
                pipeline(command("rwuniq", out_fifo,
                                 "--fields=sensor,sTime",
                                 "--bin-time=3600",
                                 "--bytes", "--packets", "--flows",
                                 "--no-titles",
                                 "--delimited=|",
                                 "--copy-input=" + flow_bin_out_path,
                                 "--output-path=" + out_file)))
    processing_pipelines = ([trend_in_rwfilter_pipeline] +
                            [trend_out_rwfilter_pipeline] +
                            trend_in_rwcount_pipelines +
                            trend_out_rwcount_pipelines)
    run_parallel(*processing_pipelines)
    # Okay, now pick up the data.
    for (bin, in_key, out_key, _, _, in_file, out_file) in bin_info:
        for l in open(in_file, 'r'):
            (sensor, time, bytes, packets, flows) = \
                (x.strip() for x in l.split('|')[:5])
            record_data(data, time, 'in', in_key, sensor,
                        bytes, packets, flows)
        for l in open(out_file, 'r'):
            (sensor, time, bytes, packets, flows) = \
                (x.strip() for x in l.split('|')[:5])
            record_data(data, time, 'out', out_key, sensor,
                        bytes, packets, flows)                
    # Clean up.  Remove the fifos and temp files.
    run_parallel(*[command("rm", "-f", filename)
                   for filename in (in_fifo, out_fifo, in_file, out_file)
                   for (_, _, _, in_fifo, out_fifo, in_file, out_file)
                       in bin_info])


########################################################################

def get_strip_size(config, default_width, default_height,
                   default_vpad, default_hpad):
    vp = config.get('chart_padding_vertical',
                    config.get('chart_padding', default_vpad))
    hp = config.get('chart_padding_horizontal',
                    config.get('chart_padding', default_hpad))
    w = config.get('chart_width', default_width)
    h = config.get('chart_strip_height', default_height)
    return (w, h, hp, vp)

def get_output_size(config, default_width, default_height,
                    default_hpad, default_vpad, default_strip_height, n):
    vp = config.get('chart_padding_vertical',
                    config.get('chart_padding', default_vpad))
    hp = config.get('chart_padding_horizontal',
                    config.get('chart_padding', default_hpad))
    w = config.get('chart_width', default_width)
    sh = config.get('chart_strip_height', default_strip_height)
    h = config.get('chart_height', default_height)
    # Calculated height based on strip size
    ch = vp + n * (sh + vp)
    if h != None and ch > h:
        # Too big, use max height
        return (w, h, hp, vp)
    else:
        # Fits, or no max height given, use calculated height
        return (w, ch, hp, vp)

def flow_param_hours(flow_params, data=None, delta=timedelta(hours=1)):
    start_time = flow_params.get_start_date()
    end_time = flow_params.get_end_date()
    time = start_time
    if time == None:
        for t in data_hours(data, delta):
            yield t
    else:
        while time <= end_time:
            yield time
            time += delta

def data_hours(data, delta=timedelta(hours=1)):
    if not data:
        return
    min_hour = max_hour = None
    for (hour, _, _) in data:
        if min_hour == None or hour < min_hour:
            min_hour = hour
        if max_hour == None or hour > max_hour:
            max_hour = hour
    time = min_hour
    while time <= max_hour:
        yield time
        time += delta
    

def full_keys(flow_params, scheme):
    hours = flow_param_hours(flow_params)
    sensors = flow_params.get_sensors()
    in_bin_keys = scheme.get_in_bin_keys()
    out_bin_keys = scheme.get_out_bin_keys()
    for time in hours:
        for sensor in sensors:
            for key in out_bin_keys:
                yield (time, 'out', key, sensor)
            for key in in_bin_keys:
                yield (time, 'in', key, sensor)

def init_full_data(flow_params, scheme):
    data = {}
    for key in full_keys(flow_params, scheme):
        data[key] = (0, 0, 0)
    return data

def record_data(data, time, direction, key, sensor, bytes, packets, flows):
    time = make_datetime(time)
    bytes = int(bytes)
    packets = int(packets)
    flows = int(flows)
    data[(time, direction, key, sensor)] = (bytes, packets, flows)

def aggregate_data(flow_params, scheme, data):
    result = {}
    for ((time, direction, key, sensor), (b, p, f)) in data.iteritems():
        (tb, tp, tf) = result.get((time, direction, key), (0, 0, 0))
        result[(time, direction, key)] = (tb + b, tp + p, tf + f)
    return result

def generate_data(in_rwfilter, out_rwfilter, flow_params, scheme, data):
    if data == None:
        data = init_full_data(flow_params, scheme)
    run_pipeline(scheme, in_rwfilter, out_rwfilter, flow_params, data)
    return data

db_create_trend_key_table = db_query("""
    create table trend_keys (
        id integer primary key,
        key varchar(1000) unique not null
    )
""")

db_insert_trend_key = db_query("""
   insert into trend_keys ( id, key )
     values ( :id, :key )
""")

db_create_trend_data_table = db_query("""
    create table trend_data (
        t timestamp not null,
        direction varchar(3) not null,
        key_id integer not null,
        sensor varchar(10) not null,
        bytes numeric not null,
        packets numeric not null,
        flows numeric not null,
        primary key ( t, direction, key_id, sensor )
    )
""")

db_delete_trend_data = db_query("""
    delete from trend_data where
        t = :t and direction = :direction and key_id = :key_id and
        sensor = :sensor
""")

db_insert_trend_data = db_query("""
    insert into trend_data ( t, direction, key_id, sensor,
                             bytes, packets, flows )
        values ( :t, :direction, :key_id, :sensor,
                 :bytes, :packets, :flows )
""")

DB_MAX_RETRIES = 6
DB_RETRY_DELAY = 30

def load_data(database_uri, database_user, database_password,
              in_rwfilter, out_rwfilter, flow_params, scheme):
    # Load all data for a given setup, regardless of whether it's been
    # seen.
    data = generate_data(in_rwfilter, out_rwfilter, flow_params, scheme, None)
    conn = db_connect(database_uri, database_user, database_password)
    # create the trend data table if it's missing
    try:
        conn.execute(db_create_trend_data_table)
        conn.commit()
    except:
        pass
    conn.rollback()
    # create the trend key table if it's missing
    try:
        conn.execute(db_create_trend_key_table)
        conn.commit()
    except:
        pass
    conn.rollback()
    # Getting/adding the keys
    try:
        retries = 0
        try:
            db_key_id = {}
            result = conn.execute(db_fetch_key_ids)
            for (key, id) in result:
                db_key_id[get_key_int_id(key)] = id
            if db_key_id:
                next_key = max(db_key_id.itervalues()) + 1
            else:
                next_key = 0
            new_keys = set(key for (_, _, key, _) in data.iterkeys()
                           if key not in db_key_id)
            for key in new_keys:
                db_key_id[key] = next_key
                next_key += 1
            for key in new_keys:
                conn.execute(db_insert_trend_key,
                             id=db_key_id[key], key=get_int_id_key(key))
            conn.commit()
        except sqlite_operationalerror:
            # try a few times
            if retries < DB_MAX_RETRIES:
                time_sleep(DB_RETRY_DELAY)
                retries += 1
                pass
            else:
                # Time's up, give up
                raise
    finally:
        conn.rollback()
    # Loading the data
    try:
        retries = 0
        try:
            for ((time, direction, key, sensor), (b, p, f)) in data.iteritems():
                conn.execute(db_delete_trend_data, t=time, direction=direction,
                             key_id=db_key_id[key],
                             sensor=sensor)
                conn.execute(db_insert_trend_data, t=time, direction=direction,
                             key_id=db_key_id[key], sensor=sensor,
                             bytes=b, packets=p, flows=f)
            conn.commit()
        except sqlite_operationalerror:
            # try a few times  
            if retries < DB_MAX_RETRIES:
                time_sleep(DB_RETRY_DELAY)
                retries += 1
                pass
            else:
                # Time's up, give up
                raise
    finally:
        conn.rollback()

def auto_load_data(database_uri, database_user, database_password,
                   in_rwfilter, out_rwfilter, flow_params, scheme):
    # Figure out what data is missing
    data = {}
    (data, missing_hours) = \
        retrieve_data(database_uri, database_user, database_password,
                      flow_params, scheme, data)
    # Throw away the data we just retrieved
    data = None
    # Load any missing data
    for new_flow_params in missing_flow_params(flow_params, missing_hours):
        load_data(database_uri, database_user, database_password,
                  in_rwfilter, out_rwfilter, new_flow_params, scheme)

db_fetch_key_ids = db_query("""
   select key, id from trend_keys
""")

db_fetch_trend_data = db_query("""
    select t, direction, key_id, sensor, bytes, packets, flows
      from trend_data
      where t >= :st and t <= :et
        and direction = 'in'
        and key_id in :in_keys
        and sensor in :sensors
    union all
    select t, direction, key_id, sensor, bytes, packets, flows
      from trend_data
      where t >= :st and t <= :et
        and direction = 'out'
        and key_id in :out_keys
        and sensor in :sensors
""")

def retrieve_data(database_uri, database_user, database_password,
                  flow_params, scheme, data):
    # Retrieve as much data as possible for a given setup, take note of
    # what is missing.
    if data == None:
        data = {}
    start_time = flow_params.get_start_date()
    end_time = flow_params.get_end_date()
    sensors = tuple(flow_params.get_sensors())
    in_bin_keys = tuple(scheme.get_in_bin_keys())
    out_bin_keys = tuple(scheme.get_out_bin_keys())
    conn = None
    try:
        try:
            conn = db_connect(database_uri, database_user, database_password)
            db_key_id = {}
            db_id_key = {}
            result = conn.execute(db_fetch_key_ids)
            for (key, id) in result:
                db_key_id[get_key_int_id(key)] = id
                db_id_key[id] = get_key_int_id(key)
            in_key_ids = tuple(db_key_id[key] for key in in_bin_keys)
            out_key_ids = tuple(db_key_id[key] for key in out_bin_keys)
            result = conn.execute(db_fetch_trend_data,
                                  st=start_time, et=end_time,
                                  sensors=sensors,
                                  in_keys=in_key_ids, out_keys=out_key_ids)
            for (time, direction, key_id, sensor,
                 bytes, packets, flows) in result:
                record_data(data, time, direction, db_id_key[key_id], sensor,
                            bytes, packets, flows)
        except:
            # If there's an error, that's okay--we'll fall back to a pull.
            pass
    finally:
        if conn != None: conn.rollback()
    # Now figure out what's missing.
    missing_hours = set()
    for (time, direction, key, sensor) in full_keys(flow_params, scheme):
        if (time, direction, key, sensor) not in data:
            missing_hours.add(time)
    return (data, missing_hours)

def missing_flow_params(flow_params, missing_hours):
    start_hour = None
    last_hour = None
    one_hour = timedelta(hours=1)
    for hour in sorted(missing_hours):
        if last_hour == None:
            start_hour = hour
            last_hour = hour
        elif last_hour == (hour - timedelta(hours=1)):
            last_hour = hour
            continue
        else:
            yield flow_params.using(start_date=start_hour, end_date=last_hour)
            start_hour = hour
            last_hour = hour
    if last_hour != None:
        yield flow_params.using(start_date=start_hour, end_date=last_hour)

def get_data(database_uri, database_user, database_password,
             in_rwfilter, out_rwfilter, flow_params, scheme):
    data = {}
    if flow_params.get_start_date():
        if (script.get_param("no-database-query") or
                script.get_param("flow-bin") or
                not database_uri):
            # Don't try to use cache.  All hours are missing.
            missing_hours = list(flow_param_hours(flow_params))
        else:
            # Try to retrieve data from the cache.  Otherwise, generate
            # data.
            (data, missing_hours) = \
                retrieve_data(database_uri, database_user, database_password,
                              flow_params, scheme, data)
        if not script.get_param("no-silk-query"):
            # Do a SiLK data pull to fill in anything that we didn't
            # already get.
            for new_flow_params in missing_flow_params(flow_params,
                                                       missing_hours):
                data = generate_data(in_rwfilter, out_rwfilter,
                                     new_flow_params, scheme, data)
        else:
            if missing_hours:
                print_warning(sys.stderr,
                              "WARNING: Some data missing but --no-silk-query "
                              "specified")
                for fp in missing_flow_params(flow_params,
                                                           missing_hours):
                    print "    %s-%s" % (
                        datetime_silk_hour(fp.get_start_date()),
                        datetime_silk_hour(fp.get_end_date()))
        full_data = init_full_data(flow_params, scheme)
        full_data.update(data)
        data = full_data
    else:
        # Not a data pull.
        data = generate_data(in_rwfilter, out_rwfilter,
                             flow_params, scheme, data)
    result = aggregate_data(flow_params, scheme, data)
    return result

def format_data(config, out, flow_params, scheme, data, values):
    bin_names = scheme.get_bin_names()
    in_bin_keys = scheme.get_in_bin_keys()
    out_bin_keys = scheme.get_out_bin_keys()
    if values == 'bytes':
        value_index = 0
    elif values == 'packets':
        value_index = 1
    else:
        value_index = 2
    out.write('|'.join(['time'] +
                       [x + '-out' for x in bin_names] +
                       [x + '-in' for x in bin_names]))
    out.write('\n')
    hours = flow_param_hours(flow_params, data)
    for time in hours:
        out.write('|'.join([datetime_silk_hour(time)] +
                           [str(data.get((time, 'out', key), (0, 0, 0))
                                        [value_index])
                            for key in out_bin_keys] +
                           [str(data.get((time, 'in', key), (0, 0, 0))
                                        [value_index])
                            for key in in_bin_keys]))
        out.write("\n")

def format_chart(config, out, flow_params, scheme, data, values, format):
    times = list(flow_param_hours(flow_params, data))
    bin_names = ["_total"] + scheme.get_bin_names()
    bin_titles = ["Total"] + scheme.get_bin_titles()
    bin_font_sizes = [14] + scheme.get_bin_title_font_sizes()
    out_bin_keys = scheme.get_out_bin_keys()
    in_bin_keys = scheme.get_in_bin_keys()
    if values == 'bytes':
        value_index = 0
        value_units = "B/s"
    elif values == 'packets':
        value_index = 1
        value_units = "p/s"
    else:
        value_index = 2
        value_units = "f/s"
    def volume_time(row):
        return "%s, %d:%02d"  % (num_prefix(row[1]/3600.0, value_units),
                                 row[0].hour, row[0].minute)
    out_total = [sum([data.get((t, 'out', key), (0, 0, 0))[value_index]
                      for key in out_bin_keys])
                 for t in times]
    out_bins = [out_total] + [[data.get((t, 'out', key),(0, 0, 0))[value_index]
                               for t in times]
                              for key in out_bin_keys]
    in_total = [sum([data.get((t, 'in', key), (0, 0, 0))[value_index]
                     for key in in_bin_keys])
                for t in times]
    in_bins = [in_total] + [[data.get((t, 'in', key), (0, 0, 0))[value_index]
                             for t in times]
                            for key in in_bin_keys]
    num_strips = len(out_bins)
    if format in ('svg', 'pdf', 'ps'):
        # Page oriented sizes
        (default_width, default_height) = (612, 792) # 8.5" x 11"
        default_strip_height = 96                    # 1 1/3"
        (default_hpad, default_vpad) = (16, 16)      # 2/9"
        title_width = 108                            # 1.5"
    else:
        # Screen oriented sizes
        (default_width, default_height) = (800, None) # 800px x unconstrained
        default_strip_height = 125                    # 125px
        (default_hpad, default_vpad) = (16, 16)       # 16px x 16px
        title_width = 125                             # 125px
    (w, h, vp, hp) = get_output_size(config, default_width, default_height,
                                     default_hpad, default_vpad,
                                     default_strip_height, num_strips)
    # Config colors
    out_color = config['chart_out_color']
    in_color = config['chart_in_color']
    chart_title = config.get('chart_title', '')
    write_chart_multi(out, format, w, h,
                      times, out_bins, in_bins, bin_titles, bin_font_sizes,
                      volume_time, chart_title, out_color=out_color,
                      in_color=in_color)
    out.close()

def format_html_chart(config, flow_params, scheme, data):
  times = list(flow_param_hours(flow_params, data))
  bin_names = ["_total"] + scheme.get_bin_names()
  bin_titles = ["Total"] + scheme.get_bin_titles()
  bin_font_sizes = [14] + scheme.get_bin_title_font_sizes()
  out_bin_keys = scheme.get_out_bin_keys()
  in_bin_keys = scheme.get_in_bin_keys()
  for output_values in ('bytes', 'packets', 'flows'):
    if output_values == 'bytes':
        value_index = 0
        value_units = "B/s"
    elif output_values == 'packets':
        value_index = 1
        value_units = "p/s"
    else:
        value_index = 2
        value_units = "f/s"
    def volume_time(row):
        return "%s, %d:%02d"  % (num_prefix(row[1]/3600.0, value_units),
                                 row[0].hour, row[0].minute)
    out_total = [sum([data.get((t, 'out', key), (0, 0, 0))[value_index]
                      for key in out_bin_keys])
                 for t in times]
    out_bins = [out_total] + [[data.get((t, 'out', key), (0, 0, 0))[value_index]
                               for t in times]
                              for key in out_bin_keys]
    in_total = [sum([data.get((t, 'in', key), (0, 0, 0))[value_index]
                     for key in in_bin_keys])
                for t in times]
    in_bins = [in_total] + [[data.get((t, 'in', key), (0, 0, 0))[value_index]
                             for t in times]
                            for key in in_bin_keys]
    (w, h, vp, hp) = get_strip_size(config, 800, 125, 16, 16)
    chart_title = config.get('chart_title', 'Trend Data')
    # Config colors
    out_color = config['chart_out_color']
    in_color = config['chart_in_color']
    if output_values == 'bytes':
        output_filename = 'index.html'
    else:
        output_filename = 'index-%s.html' % output_values
    html_out = script.get_output_dir_file('html-dir-path',
                                          output_filename,
                                          mime_type='text/html')
    html_out.write("""<?xml version="1.0"?>
<html>
  <head>
    <title>%(chart_title)s</title>
    <style type="text/css">
      html {
        font-family: sans-serif;
      }
      body {
        width: 800px;
        padding: 16px;
        margin: 0 auto;
        position: relative;
      }
      h1 {
        padding-right: 166px;
      }
      ul.l-nav {
        display: block; margin: 0; padding: 0;
      }
      ul.l-nav li {
        display: inline; margin: 0; padding: 0; padding-right: 0.5em;
        font-weight: bold;
      }
      ul.l-nav a {
        font-weight: normal;
      }
      .l-scale {
        position: absolute;
        top: 16px;
        right: 16px;
        width: 150px;
        font-size: 10px;
      }
      .l-scale p {
        margin: 0;
      }
    </style>
  </head>
  <body>
    <h1>%(chart_title)s (%(output_values)s)</h1>
    <ul class="l-nav">""" % {
            'chart_title': chart_title,
            'output_values': output_values})
    for ov in ('bytes', 'packets', 'flows'):
        if ov == output_values:
            html_out.write("""
      <li>%(ov)s</li>""" % {'ov': ov})
        else:
            if ov == 'bytes':
                html_out.write("""
      <li><a href="index.html">%(ov)s</a></li>""" % {'ov': ov})
            else:
                html_out.write("""
      <li><a href="index-%(ov)s.html">%(ov)s</a></li>""" % {'ov': ov})
    html_out.write("""
    </ul>
    <div class="l-scale">
      <p>m%(value_units)s = 10<sup>-3</sup> %(output_values)s per second</p>
      <p>k%(value_units)s = 10<sup>3</sup> %(output_values)s per second</p>
      <p>M%(value_units)s = 10<sup>6</sup> %(output_values)s per second</p>
      <p>G%(value_units)s = 10<sup>9</sup> %(output_values)s per second</p>
    </div>
""" % {'value_units': value_units, 'output_values': output_values})
    for (i, name) in enumerate(bin_names):
        html_out.write("""
    <img src="%(name)s-%(output_values)s.png" alt="data" style="width:%(width)dpx; height:%(height)dpx"/>""" %
                       {'title': bin_titles[i], 'name': name,
                        'output_values': output_values,
                        'width': w, 'height': h})
    html_out.write("""
  </body>
</html>
""")
    html_out.close()
    for (i, name) in enumerate(bin_names):
        out = script.get_output_dir_file('html-dir-path',
                                         '%s-%s.png' % (name, output_values),
                                         mime_type="image/png")
        write_chart_single(out, 'png', w, h, times, out_bins[i], in_bins[i],
                           bin_titles[i], bin_font_sizes[i], volume_time,
                           out_color=out_color, in_color=in_color)
        out.close()

def main():
    try:
        config = read_config(script.get_param('config-path'))
    except TrendConfigError:
        raise script.UserError(str(sys.exc_info()[1]))
    scheme = config['trend_scheme']

    flow_bin = script.get_param('flow-bin')
    if flow_bin:
        if flow_bin not in scheme.get_bin_names():
            raise script.ParamError(
                "flow-bin", flow_bin,
                "Bin name not defined in config file")
        if not (script.get_param('in-flow-path') or
                script.get_param('out-flow-path')):
            raise script.ParamError(
                "flow-bin", flow_bin,
                "'flow-bin' given, but neither 'in-flow-path' nor "
                "'out-flow-path' is set")
    # Check chart params.
    if not (script.get_param('data-path') or
            script.get_param('png-path') or
            script.get_param('svg-path') or
            script.get_param('pdf-path') or
            script.get_param('ps-path') or
            script.get_param('html-dir-path') or
            script.get_param('load-database') or
            script.get_param('auto-load-database') or
            script.get_param('flow-bin')):
        raise script.UserError("No output specified.")
    if script.get_param('chart-title') != None:
        config['chart_title'] = script.get_param('chart-title')
    output_values = script.get_param('values')
    if output_values not in ('bytes', 'packets', 'flows'):
        raise script.ParamError(
            "values", output_values,
            "must be 'bytes', 'packets', or 'flows'")
    used_rwfilter = False
    database_uri = config['database_uri']
    database_user = script.get_param('database-user')
    database_password = script.get_param('database-password')
    in_rwfilter = config["in_rwfilter"]
    if in_rwfilter:
        in_rwfilter = shlex.split(in_rwfilter)
        used_rwfilter = True
    else:
        in_types = script.get_param("in-types")
        if in_types == None:
            in_types = config.get('in_types', None) or 'in,inweb'
        in_rwfilter = ["rwfilter", "--type=" + in_types,
                       "--compression-method=none",
                       "--all-destination=stdout"]
    out_rwfilter = config["out_rwfilter"]
    if out_rwfilter:
        if not used_rwfilter:
            raise script.UserError(
                "in-filter and out-filter must be given together in "
                "config file")
        out_rwfilter = shlex.split(out_rwfilter)
    else:
        if used_rwfilter:
            raise script.UserError(
                "in-filter and out-filter must be given together in "
                "config file")
        out_types = script.get_param("out-types")
        if out_types == None:
            out_types = config.get('out_types', None) or 'out,outweb'
        out_rwfilter = ["rwfilter", "--type=" + out_types,
                       "--compression-method=none",
                        "--all-destination=stdout"]
    load_database = (script.get_param('load-database') or
                     script.get_param('auto-load-database'))
    flow_params = script.get_flow_params()
    if load_database:
        if database_uri == None:
            raise script.ParamError(
                "load-database", None,
                "Config file must define database_uri")
        if script.get_param('auto-load-database'):
            auto_load_data(database_uri, database_user, database_password,
                           in_rwfilter, out_rwfilter, flow_params, scheme)
        else:
            load_data(database_uri, database_user, database_password,
                      in_rwfilter, out_rwfilter, flow_params, scheme)
    data = get_data(database_uri, database_user, database_password,
                    in_rwfilter, out_rwfilter, flow_params, scheme)
    data_out = script.get_output_file('data-path')
    if data_out != None:
        format_data(config, data_out, flow_params, scheme, data, output_values)
    if chart_support:
        png_out = script.get_output_file('png-path')
        if png_out != None:
            format_chart(config, png_out, flow_params, scheme, data,
                         output_values, 'png')
        svg_out = script.get_output_file('svg-path')
        if svg_out != None:
            format_chart(config, svg_out, flow_params, scheme, data,
                         output_values, 'svg')
        if pdf_support:
            pdf_out = script.get_output_file('pdf-path')
            if pdf_out != None:
                format_chart(config, pdf_out, flow_params, scheme, data,
                             output_values, 'pdf')
        ps_out = script.get_output_file('ps-path')
        if ps_out != None:
            format_chart(config, ps_out, flow_params, scheme, data,
                         output_values, 'ps')
        html_out = script.get_param('html-dir-path')
        if html_out != None:
            if os.path.exists(html_out):
                if not os.path.isdir(html_out):
                    raise script.UserError(
                        "html-dir-path %s is not a directory" %
                        repr(html_out))
            else:
                os.mkdir(html_out)
            format_html_chart(config, flow_params, scheme, data)

script.execute(main)
